# File Maker for Photo Color Adjustment 16/8bit Version
# Ver. 4.51a
# add psedo flat-field correction algorism
# Silent & Verbose Version
# (c)2020-22 Ohnishi, Yasuo
from ij import IJ, ImagePlus 
from ij import plugin
from ij import WindowManager as WM
from ij.io import DirectoryChooser
from ij.io import OpenDialog
from ij.gui import GenericDialog
from ij.plugin import ChannelSplitter
from ij.plugin import Duplicator
from ij.process import ImageConverter
from loci.plugins import BF
import os
import re
#import sys
#import io
import datetime
from org.python.core import codecs
codecs.setDefaultEncoding('utf-8')
def convValue(Value, Mode):
#   Converting Luminance Value each in 8, 16, 32bit mode
    if Mode == "16-bit":
        Value = Value * 256
    elif Mode == "32-bit":
        Value = Value / 255
    else:
        Value = Value
    return Value
############### 2022/7/5
def make_Gaussian_imp(imp_org, sigma):
    imp_width = imp_org.getWidth()
    imp_height  = imp_org.getHeight()
    swidth = int(imp_width / 4)
    sheight = int(imp_height / 4)
    imp_g = Duplicator().run(imp_org)  
    ImageConverter(imp_g).convertToGray16()
        # imp_g will be gauss image of level adjusted imp_org
    imp_g.resize(swidth, sheight, 'bilinear')
    IJ.run(imp_g, "Gaussian Blur...", "sigma=" + str(sigma))
    imp_g.resize(imp_width, imp_height, 'bilinear')
    return imp_g # imp_g will be 16-bit
#----------------
def psedo_Flat_field_C(imp_g0, imp_b0, sigma, Mode, folder, \
        filename, parm_subtractBv, parm_leveladj, parm_correct_fac):
    """
        imp_g0: reference ch. imp, imp_b0: B ch. imp
        sigma: Gaussian blur sigma
        parm_subtractBv: Subtraction or not B when B is much higher than G
        parm_leveladj: adjust or not luminance level
        parm_correct_fac: manual correction factor
    """
### Options for developping
    if parm_subtractBv == False: # No Subtraction B when B is much higher than G
        test = "A"  # select test "A" or "B"
    else:  # Subtraction B when B is much higher than G
        test = "B"  # select test "A" or "B"
    writingout = False  # select writingout True or False
    if parm_leveladj != "No":
        meanadjustment = True # select mean adjustment True or False
    else:
        meanadjustment = False # select mean adjustment True or False
    ####
    imp_g1 = make_Gaussian_imp(imp_g0, sigma)
    imp_b1 = make_Gaussian_imp(imp_b0, sigma)

    """
        imp_g1: G ch. Gaussan blur image
        imp_b1: B ch. Gaussan blur image
    """
    if meanadjustment == True: # adjust level difference of G and B Ver. 3.53
        stat_g1 = imp_g1.getStatistics()
        stat_b1 = imp_b1.getStatistics()
        if parm_leveladj == "Auto":
            if stat_g1.mean < stat_b1.mean: 
                  # adjust level difference of G and B Gaussian image (Ver. 3.52)
                IJ.run(imp_g1, "Add...", "value-" + str(stat_g1.mean - stat_b1.mean))
        elif parm_leveladj == "Manual": # Manual adjustment of correction amount of B
            IJ.run(imp_g1, "Add...", "value-" + str(convValue(parm_correct_fac, Mode)))
                # add correction value "parm_correct_fac" 
    ic = plugin.ImageCalculator()
    imp_dif_gb = ic.run("Subtract create", imp_g1, imp_b1) 
                # imp_dif_gb: darker region of Gaussian B than G
    imp_c_b0 = ic.run("Add create", imp_b0, imp_dif_gb) 
                # image correction of B (subtract yellowed value from B)
    if test =="B":
        # subtract higher region of B than G
        imp_dif_bg = ic.run("Subtract create", imp_b1, imp_g1) # No window made
        imp_c_b = ic.run("Subtract create", imp_c_b0, imp_dif_bg) # No window made
        imp_dif_bg.close()
    else:
        imp_c_b = imp_c_b0 # for test "A"
    stat_c_b = imp_c_b.getStatistics()
    """
    if stat_c_b.mean > stat_g.mean:
        IJ.run(imp_c_b, "Subtract...", "value=" + str(int(stat_c_b.mean - stat_g.mean)))
         not good!
    """
    if writingout == True:
        IJ.saveAsTiff(imp_dif_gb, folder + filename +"_imp_Gaussian-dif_gb.tif")
        if test =="B":
            IJ.saveAsTiff(imp_dif_bg, folder + filename +"_imp_Gaussian-dif_bg.tif")
    if Mode == "8-bit":
        ImageConverter(imp_c_b).convertToGray8()
    elif Mode == "16-bit":
        ImageConverter(imp_c_b).convertToGray16()
    imp_dif_gb.close()
    imp_c_b0.close()
    return imp_c_b, imp_b1, imp_g1

#---mask for excluding more bluer region from correction (for Hybrid correction)
def make_fg_mask1(impG, impB, folder, filename):
    """
    This routine is making mask for foreground layer 
    """
    debugmode = False
    ic = plugin.ImageCalculator()
    imp8_B = Duplicator().run(impB) # corrected B ch.
    imp8_G = Duplicator().run(impG) # G ch.
    statB = impB.getStatistics()
    statG = impG.getStatistics()
    #IJ.showMessage("imp8_cB Mean: " + str(stat8_cB.mean / 256) + " G Mean: " + str(statG.mean / 256))
    if statG.mean < statB.mean: #adjust B value relative to G
        IJ.run(imp8_G, "Add...", "value=" + str(int(statB.mean - statG.mean)))
    imp8_B_G = ic.run("Subtract create", impB, imp8_G)
    imp8_G.close()
        # extract more green region 
    stat = imp8_B_G.getStatistics()
    if debugmode == True:
        IJ.saveAsTiff(imp8_B_G, folder + filename +"_R+128_imp_B_G.tif")
        IJ.showMessage("imp8_B_G Max: " + str(stat.max / 256) + " Mean: "\
                                                + str(stat.mean / 256))
    IJ.run(imp8_B_G, "Multiply...", "value=7")
    IJ.run(imp8_B_G, "Invert", "") # bluer region  > blacker (opaque)
                                   # not to apply correction in bluer region
    imp8_0 = imp8_B_G
    IJ.run(imp8_0, "Gamma...", "value=0.5")
    IJ.run(imp8_0, "Gaussian Blur...", "sigma=1")
    imp8_B.close()
    imp8_G.close()
    return imp8_0
    """
    imp8_0:
        region bluer > more opaque (not to apply correction)
    """
#---mask for preserving plants green region from correction (for psedo FF correction)
def make_fg_mask2(impG, imp_c_b, imp_R128, impB, impR, Mode, parm_Binary, folder, filename):
    """
    This routine is making mask for foreground layer in psedo flat-field 
                                                          correction algorism
    impG: G ch. / imp_c_b: corrected B ch. 
    imp_R128: mask for background correction mask (R128)
    """
    debugmode = False
    ic = plugin.ImageCalculator()
    imp8_cB = Duplicator().run(imp_c_b) # corrected B ch.
    imp8_G = Duplicator().run(impG) # G ch.
    imp8_R = Duplicator().run(impR) # R ch.

    stat8_cB = imp8_cB.getStatistics()
    statG = impG.getStatistics()
    """
    IJ.showMessage("imp8_cB Mean: " + str(stat8_cB.mean / 256) + " G Mean: " + \
                                                           str(statG.mean / 256))
    """
    if statG.mean < stat8_cB.mean: #adjust B value relative to G
        IJ.run(imp8_G, "Add...", "value=" + str(int(stat8_cB.mean - statG.mean)))
    # IJ.run(imp8_G, "Subtract...", "value=" + str(convValue(5, Mode))) # limit the Green Region
    # --make G > B image--
    imp8_G_cB = ic.run("Subtract create", imp8_G, imp8_cB)
        # extract darker image of G > B / att! darker parts will be lighter
    imp8_G.close()
        # extract more green region 
    stat = imp8_G_cB.getStatistics()
    if debugmode == True:
        IJ.saveAsTiff(imp8_cB, folder + filename +"_R+128_imp_c_b.tif")
        IJ.saveAsTiff(imp8_G_cB, folder + filename +"_R+128_imp8_G_cB.tif")
        IJ.showMessage("imp8_G_cB Max: " + str(stat.max / 256) + " Mean: "\
                                                + str(stat.mean / 256))
    IJ.run(imp8_G_cB, "Multiply...", "value=7") # value 10 ok?
    IJ.run(imp8_G_cB, "Invert", "") # more green region  > more black (opaque)
    if debugmode == True:
        IJ.saveAsTiff(imp8_G_cB, folder + filename +"_R+128_imp8_G_cB_inverted.tif")
    #- relative green region
    IJ.run(imp8_cB, "Multiply...", "value=0.5") 
    IJ.run(imp8_R, "Multiply...", "value=0.5") 
    imp8_RcB = ic.run("Add create", imp8_cB, imp8_R) 
    imp8_rG = ic.run("Subtract create", imp8_RcB, imp8_G) 
        # imp8_rG: relative not green region will be lighter
    imp8_rG.getProcessor().threshold(int(convValue(parm_Binary, Mode))) # Make binary Mask
    IJ.run(imp8_rG, "Multiply...", "value=" + str(convValue(1, Mode))) # 16, 32bit image > Multiply
    imp8_rG.updateAndDraw()

    imp8_2 = ic.run("Max create", imp8_rG, imp8_G_cB) 
    # IJ.run(imp_R128, "Invert...","")
    imp8_0 = ic.run("Max create", imp8_2, imp_R128) 
#    IJ.run(imp8, "Gamma...", "value=0.5")
#    IJ.run(imp8, "Gaussian Blur...", "sigma=20")
    imp8_2.close()
    imp8_cB.close()
    imp8_R.close()
    imp8_G_cB.close()
    imp8_RcB.close()
    imp8_rG.close()
    return imp8_0
    """
    imp8_0:
        region relatively green > more opaque
    """
#-----------------------------------------
def RGBmanipulation(folder,filename,parm):
#  input Parameters
    if parm is not None:
 # if parm is not null object, then...
        Mode, parm_Binary, parm_Gain, parm_Shadow, Show, parm_Prph, parm_FGGain,\
            parm_FFCorrection, parm_FFC_sigma, parm_Gmixrate, parm_subtractBv,\
            parm_leveladj, parm_correct_fac, HybridC, Fg_mask_type = parm
        print(Mode, parm_Binary, parm_Gain, parm_Shadow, Show, parm_Prph, parm_FGGain) 
#     Mode: Image bit depth 16-bit or 8-bit
#     Show: Show working windows or not
#     parm_Binary: Threshold of fore / background
#     parm_Gain: Lightness of Distance adjusting Layer
#     parm_Shadow: Threshold of Dark part adjustment image
#     parm_Prph: Threshold of Periphery adjustment image
#     parm_FGGain: Threshold of Foreground mask

# Always Open file with Bio-format Plugin
    imps = BF.openImagePlus(folder + filename) # Open file with Bio-format Plugin
    imp0 = imps[0] #imp0: Original Image 
    if Mode == "Auto":
        bitdepth = imp0.getBitDepth()
        if bitdepth == 8:
            Mode = "8-bit"
        elif bitdepth ==16:
            Mode = "16-bit"
        elif bitdepth ==32:
            Mode = "32-bit"
        else:
            Mode = "16-bit"
    if Show == "Yes":
 # Windows Showing Mode
        imp0.show()
#------
    impsCh = ChannelSplitter.split(imp0) #Channel Splitting of imp0
    for i in impsCh: # it needs to convert Grayscale image
        if Mode == "8-bit":
            ImageConverter(i).convertToGray8()
        elif Mode == "16-bit":
            ImageConverter(i).convertToGray16()
        else:
            ImageConverter(i).convertToGray32()
## R ch.
    impR =  impsCh[0]
    IJ.saveAsTiff(impR, folder + filename +"_R.tif")
## G ch.
    impG =  impsCh[1]
    IJ.saveAsTiff(impG, folder + filename +"_G.tif")
## B ch. & B x 0.3
    impB =  impsCh[2]
    IJ.saveAsTiff(impB, folder + filename +"_B.tif")
# Show Windows
    imp0.close()
    if Show == "Yes":
        impR.show()
        impG.show()
        impB.show()
#--- preparetion for making layer images / to make (R+G)/2 image
#         & periphery correction layer
# G ch. x 0.5
#        Make impG05  as G x 0.5 Image
    impG05 = Duplicator().run(impG)
    if Show == "Yes":
        impG05.show()
    IJ.run(impG05, "Multiply...", "value=0.5")
# R ch. x 0.5
#        Make impR05  as R x 0.5 Image
    impR05 = Duplicator().run(impR)
    if Show == "Yes":
        impR05.show()
    IJ.run(impR05, "Multiply...", "value=0.5")
# B x 0.5 + R x 0.5
    ic = plugin.ImageCalculator()
    imp5 = ic.run("Add create", impR05, impG05) # No window made
    #-----save periphery correction layer (no psedo FF correction)
    if (parm_FFCorrection == False) and (HybridC == "Off"):
        IJ.saveAsTiff(imp5, folder + filename +"_G+R.tif")
    #-----save periphery correction layer (psedo FF correction)
    else:
        imp_periph, imp_Gauss_b1, imp_Gauss_rg1 = \
            psedo_Flat_field_C(imp5, impB, parm_FFC_sigma, Mode, folder,\
            filename, True, parm_leveladj, parm_correct_fac) 
           # reference image must be impG or imp5?
        IJ.saveAsTiff(imp_periph, folder + filename +"_G+R.tif")
    if Show == "Yes": #Hide the image windows to close
        imp5.show()
        impR05.hide()
        impG05.hide()
    impR05.close()
    impG05.close()

#--- making foreground layer image
#  executing psedo Flat-Field correction
#    HybridC = "On"  # Off, On: Normal
    if HybridC != "Off":
        parm_leveladj = "Auto"
#   always execute psedo_Flat_field_C for mask
    imp_c_b, imp_Gauss_b1, imp_Gauss_g1 = \
        psedo_Flat_field_C(impG, impB, parm_FFC_sigma, Mode, folder,\
        filename, parm_subtractBv, parm_leveladj, parm_correct_fac) 
        #-----save periphery correction layer (apply psedo FF correction)
##      Make impB03 as B x 0.3 Image > B:G ratio variable
    if parm_FFCorrection == True: # new algorithm
        impB03 = Duplicator().run(imp_c_b)
        Bmixrate = 1.0 - parm_Gmixrate
        strValueB =  "value=" + str(Bmixrate)
        strValueG =  "value=" + str(parm_Gmixrate)
    elif HybridC != "Off": # new algorithm
        impB03 = Duplicator().run(imp_c_b)
        strValueB =  "value=0.3"
        strValueG =  "value=0.7"
    else: # existing algorithm
        impB03 = Duplicator().run(impB)
        strValueB =  "value=0.3"
        strValueG =  "value=0.7"
    statB03 = impB03.getStatistics()
    IJ.run(impB03, "Multiply...", strValueB)
    if Show == "Yes":
        impB03.show()
# G ch. x 0.7
##        Make impG07 as G x 0.7 Image
    impG07 = Duplicator().run(impG)
    statG07 = impG07.getStatistics()
    if statG07.mean < statB03.mean:
        IJ.run(impG07, "Add...", "value=" + str(statB03.mean - statG07.mean))
        # adjust G value for mixing to B
    IJ.run(impG07, "Multiply...", strValueG)
    if Show == "Yes":
        impG07.show()
# B x 0.3 + G x 0.7
    ic = plugin.ImageCalculator()
    imp3 = ic.run("Add create", impB03, impG07) # No window made
    IJ.saveAsTiff(imp3, folder + filename +"_B_foreground.tif")
        # even in new alogorithm G mixing is needed to improve texture
    imp3.close()
#--- making background layer image
# (G ch.+30) x 0.7
##        Make impG30x07 as (G+30) x 0.7 Image
    if parm_FFCorrection == False:  # no applying Flat-field
        impG30x07 = Duplicator().run(impG)
        if Show == "Yes":
            impG30x07.show()
        # Gain for background layer image
        # appropriate gamma is /1.45(= 0.69) for perceptive TRC image
        if parm_FGGain == 128: # in case of perceptive gamma TRC image.
            IJ.run(impG30x07, "Gamma...", "value=" + str(0.69))
        else: # in case of linear image.
            IJ.run(impG30x07, "Gamma...", "value=" + str(0.7))
        IJ.run(impG30x07, "Multiply...", "value=0.7")
    ## B x 0.3 + (G+30) x 0.7 + x(parm_Gain)
        ic = plugin.ImageCalculator()
        imp4 = ic.run("Add create", impB03, impG30x07) # No window made
    #  add B x 0.3 + (G+30) x 0.7 + x
        IJ.run(imp4, "Add...", "value=" + str(convValue(parm_Gain, Mode))) 
                                        # add Value parm_Gain
        if Show == "Yes":
  #Hide the image windows to close
            impG30x07.hide()
        impG30x07.close()
    else:  # applying Flat-field
        imp4 = Duplicator().run(imp_c_b)
    IJ.saveAsTiff(imp4, folder + filename +"_B_Background.tif")
    imp4.close()
    if Show == "Yes":
        impG07.hide()
        impB03.hide()
    impG07.close()
    impB03.close()
#--- making Periphery Adjustment Mask
# Periphery Adjustment Mask: B - ((R+G)/2-10)
    impGR = Duplicator().run(imp5) #imp5: (G+R)/2 image
    if Show == "Yes":
        impGR.show()
#    if parm_FGGain != 128:
#        IJ.run(impB, "Gamma...", "value=" + str(1/2.2))
#        IJ.run(impGR, "Gamma...", "value=" + str(1/2.2))
    IJ.run(impGR, "Subtract...", "value="+ str(convValue(parm_Prph, Mode))) 
                                               # subtract Value parm_Prph /10
    ic = plugin.ImageCalculator()
    if parm_FFCorrection == False:
        imp2GR = ic.run("Subtract create", impB, impGR) #impB - impGR:(G+R/2)
    else:
        imp2GR = ic.run("Subtract create", imp_c_b, impGR) #impB - impGR:(G+R/2)
    if Show == "Yes":
        imp2GR.show()
    IJ.run(imp2GR, "Multiply...", "value=4") # Save the Subtracted image x 4
    IJ.saveAsTiff(imp2GR, folder + filename +"_Periphery_Adjst_Mask.tif") 
    if Show == "Yes":
 #Hide the image windows to close
        imp2GR.hide()
        impGR.hide()
    imp2GR.close()
 #Close the images
    impGR.close()
#
#--- making dark parts correction layer 
# Dark Parts 0 - parm_Shadow
    imp6 = Duplicator().run(imp5) #imp5: (G+R)/2 image, imp6: dark part image
    if Show == "Yes":
        imp6.hide()
    IJ.run(imp6, "Add...", "value=" + str(convValue(255 - parm_Shadow, Mode))) 
    IJ.run(imp6, "Subtract...", "value=" + str(convValue(255 - parm_Shadow, Mode)))
    IJ.saveAsTiff(imp6, folder + filename +"_Dark.tif")
    if Show == "Yes":
 #Hide the image windows to close
        imp5.hide()
        imp6.hide()
    imp5.close()
 #Close the images
    imp6.close()
#--- making background correction mask 
# Make Binary Mask: Threshold R parm_Binary
    imp9 = Duplicator().run(impR)
    if Show == "Yes":
        imp9.show()
    imp9.getProcessor().threshold(int(convValue(parm_Binary, Mode))) # Make binary Mask
    if Mode != "8-bit":
        IJ.run(imp9, "Multiply...", "value=" + str(convValue(1, Mode))) # 16, 32bit image > Multiply 256 ^ 1 or 2
        IJ.run(imp9, "Add...", "value=" + str(convValue(1, Mode) - 1)) # if 16-bit > make Max Value 65535
    imp9.updateAndDraw()
    IJ.saveAsTiff(imp9, folder + filename +"_Binary_Mask.tif")
#--- making foreground correction mask 
# R + parm_FGGain(128) ## mask image for foreground correction layer
#---- traditional R+128 mask
    if Fg_mask_type == "Type1":  # Tradisional
        imp8 = Duplicator().run(impR)
        IJ.run(imp8, "Add...", "value=" + str(convValue(parm_FGGain, Mode)))
        if Show == "Yes":
            imp8.show()
        IJ.saveAsTiff(imp8, folder + filename +"_R+128.tif")
        if Show == "Yes":
            imp8.hide()
        imp8.close()
#---- mask for applying correction in region impB is darker than impG (in Hybrid Correction)
    elif (Fg_mask_type == "Type2") or (Fg_mask_type == "Type4"):  # applying Hybirid correction
        ic = plugin.ImageCalculator()
        imp8a = make_fg_mask1(impG, impB, folder, filename)
        if Fg_mask_type == "Type4":
            imp8b = make_fg_mask2\
                 (impG, imp_c_b, imp9, impB, impR, Mode, parm_Binary, folder, filename)
            imp8 = ic.run("Min create", imp8a, imp8b)
        else:
            imp8 = imp8a
        # IJ.run(imp8, "Multiply...", "value=" + str(convValue(1, Mode))) # 16, 32bit image > Multiply 256 ^ 1 or 2
        imp8.updateAndDraw()
#---- mask for adjustment plants green in applying Flat-field correction
    else:   # applying Flat-field correction
#        imp8 = make_Gaussian_fg_mask(impG, imp_c_b, imp9, folder, filename)
        imp8 = make_fg_mask2\
                      (impG, imp_c_b, imp9, impB, impR, Mode, parm_Binary, folder, filename)
    if Show == "Yes":
        imp8.show()
    IJ.saveAsTiff(imp8, folder + filename +"_R+128.tif")
    if Show == "Yes":
        imp8.hide()
    imp8.close()
#---------------------
    if Show == "Yes":
 #Hide the imp9 window to close
        imp9.hide()
        impB.hide()
    imp9.close()
    impR.close()
    impG.close()
    impB.close()
#------------------------------------------------
# Novice Mode: Get parameters from the input DialogBox 
def NoviceMenu(path):
    gd = GenericDialog("Novice Menu: Enter Parameters")
    TRCType = ["Perceptual", "Linear"]
    gd.addChoice("Image TRC Type", TRCType, TRCType[0])
    gd.addNumericField("Background Mask Threshold (0 to 255)",128, 0)
    # gd.addNumericField("Background Image Gain (Max to 100)", 0, 0)
    gd.addNumericField("Shadow Area Threshold (0 to 255)", 50, 0)
    gd.addNumericField("Periphery Adjst Threshold (-10 to 20)", 10, 0)
    HybridC_Choice = ["Off", "On"]
    gd.addChoice("Apply hybrid correction?", HybridC_Choice, HybridC_Choice[1])
    gd.addCheckbox("(in Hybrid correction) Preserve plants green", False)
    showWin = ["Yes","No"]
    gd.addChoice("Show working windows?", showWin, showWin[0])
    gd.addMessage("If you would like to use Expert Menu, Please cancel this menu.")
    gd.addMessage("Recommended Parameters for Linear Images\n 60 for Background Mask Threshold\n 3 for Shadow Area Threshold\n 0 for Periphery Adjst Threshold")
    gd.showDialog()
    parm_Mode   = "Auto"
    TRCselected   = gd.getNextChoice()
    if TRCselected == "Perceptual":
        parm_FGGain = 128
    else:
        parm_FGGain = 56
    parm_Binary   = gd.getNextNumber()
#    parm_Gain = gd.getNextNumber()
    parm_Shadow = gd.getNextNumber()
    parm_Prph = gd.getNextNumber()
    parm_HybridC   = gd.getNextChoice()
    green_preserve = gd.getNextBoolean()
    parm_Show   = gd.getNextChoice()
    parm_Gain = 0.0
    parm_FFCorrection = False
    parm_FFC_sigma = 20.0
    parm_Gmixrate = 0.0
    parm_leveladj  = "Auto"
    parm_correct_fac = 0.0
    parm_subtractBv = False
    if parm_Binary > 255: #Set Max Value
        parm_Binary = 255
    if parm_Gain > 100:
        parm_Binary = 100
    if parm_Shadow > 255: #Set Max Value
        parm_Shadow = 255
    if parm_HybridC == "On" and green_preserve == True:
        Fg_mask_type = "Type4"
    elif parm_HybridC == "On" and green_preserve == False:
        Fg_mask_type = "Type2"
    else:
        Fg_mask_type = "Type1"
    if gd.wasCanceled():  
        parm_Mode   = "Expert"
        IJ.log("Expert Menu called")
        return str(parm_Mode), parm_Binary, parm_Gain, parm_Shadow,str(parm_Show), \
            parm_Prph, parm_FGGain, parm_FFCorrection, parm_FFC_sigma, parm_Gmixrate, \
            parm_subtractBv, parm_leveladj, parm_correct_fac, parm_HybridC, Fg_mask_type
    write_out(path, parm_Binary, parm_Gain, parm_Shadow, parm_Prph, \
            parm_FGGain, parm_FFCorrection, parm_FFC_sigma, parm_Gmixrate, \
            parm_subtractBv, parm_leveladj, parm_correct_fac, parm_HybridC, Fg_mask_type)
    return str(parm_Mode), parm_Binary, parm_Gain, parm_Shadow, str(parm_Show), \
        parm_Prph, parm_FGGain, parm_FFCorrection, parm_FFC_sigma, parm_Gmixrate, \
        parm_subtractBv, parm_leveladj, parm_correct_fac, parm_HybridC, Fg_mask_type
#----------------------------------------------------
# Expert Mode: Get parameters from the input DialogBox 
def getParameters(path):
    oldparms = read_in(path)
    """
    print "oldparms"
    print oldparms["parm_Binary"]
    print oldparms["parm_FGGain"]
    """
    gd = GenericDialog("Expert Menu: Enter Parameters")
    TRCType = ["Perceptual", "Linear"]
    """
    # IJ.showMessage["parm_FGGain" + str(oldparms["parm_FGGain"])]
    """
    if int(oldparms["parm_FGGain"]) == 128:
        gd.addChoice("Image TRC Type", TRCType, TRCType[0])
    else:
        gd.addChoice("Image TRC Type", TRCType, TRCType[1])

    gd.addNumericField("Background Mask Threshold (0 to 255)",float(oldparms["parm_Binary"]), 0)
    gd.addNumericField("Background Image Gain (Max to 100)", float(oldparms["parm_Gain"]), 0)
    gd.addNumericField("Shadow Area Threshold (0 to 255)", float(oldparms["parm_Shadow"]), 0)
    gd.addNumericField("Periphery Adjst Threshold (-10 to 20)", float(oldparms["parm_Prph"]), 0)

    HybridC_Choice = ["Off", "On"]
    gd.addChoice("Apply hybrid correction?", HybridC_Choice, HybridC_Choice[0])
    #gd.addCheckbox("(in Hybrid correction) Preserve plants green", eval(oldparms["green_preserve"]) )
    mask_type = ["Type1", "Type2", "Type3", "Type4"]
    gd.addChoice("Mask type for foreground correction layer", mask_type, mask_type[1])
    gd.addMessage("       Type1: R+128 mask " + \
          "\n       Type2: Apply darker B region " \
        + "\n       Type3: Preserve plants green \n       Type4: Type2 and 3")
# Type1: traditional / Type2: B > G / Type3: Green preserve / Type4: Type2 & 3 
    gd.addMessage("-----------------------------------")
    gd.addMessage("[Extended Psedo Flat-Field correction parameters]")
    gd.addCheckbox("apply psedo Flat-Field correction", eval(oldparms["parm_FFCorrection"]) )
    gd.addNumericField("Sigma for Flat-Field correction", float(oldparms["parm_FFC_sigma"]), 1)
    gd.addNumericField("G mix rate to B (with FF Correction)", float(oldparms["parm_Gmixrate"]), 1)
    leveladj = ["No","Auto","Manual"]
    if oldparms["parm_leveladj"] == "Auto":
        gd.addChoice("Correction Amount adjustment", leveladj, leveladj[1])
    elif oldparms["parm_leveladj"] == "Manual":
        gd.addChoice("Correction Amount adjustment", leveladj, leveladj[2])
    else:
        gd.addChoice("Correction Amount adjustment", leveladj, leveladj[0])
    gd.addNumericField("Incremental Amount (in Manual 0-255)", \
                                         float(oldparms["parm_correct_fac"]), 0)
    gd.addCheckbox("Subtract Pixel Value if B > G",\
                                              eval(oldparms["parm_subtractBv"]))
    gd.addMessage("----------")
    showWin = ["Yes","No"]
    gd.addChoice("Show working windows?", showWin, showWin[0])
    gd.addMessage("Recommended Parameters for Linear Images\n 60 for Background Mask Threshold\n 3 for Shadow Area Threshold\n 0 for Periphery Adjst Threshold")
    gd.showDialog()
    parm_Mode   = "Auto"
    TRCselected   = gd.getNextChoice()
    if TRCselected == "Perceptual":
        parm_FGGain = 128
    else:
        parm_FGGain = 56
    parm_Binary   = gd.getNextNumber()
    parm_Gain = gd.getNextNumber()
    parm_Shadow = gd.getNextNumber()
    parm_Prph = gd.getNextNumber()
    parm_HybridC   = gd.getNextChoice()
#    green_preserve = gd.getNextBoolean()
    Fg_mask_type  = gd.getNextChoice()
    parm_FFCorrection = gd.getNextBoolean() #get value of checkbox
    parm_FFC_sigma = gd.getNextNumber()
    parm_Gmixrate = gd.getNextNumber()
    parm_leveladj  = gd.getNextChoice()
    parm_correct_fac = gd.getNextNumber()
    parm_subtractBv = gd.getNextBoolean() #get value of checkbox
    parm_Show   = gd.getNextChoice()

    """
         parm_leveladj, parm_correct_fac
    """
    if parm_Binary > 255: #Set Max Value
        parm_Binary = 255
    if parm_Gain > 100:
        parm_Binary = 100
    if parm_Shadow > 255: #Set Max Value
        parm_Shadow = 255
        if parm_Gmixrate > 1.0:
            parm_Gmixrate = 1.0
        elif parm_Gmixrate < 0.0:
            parm_Gmixrate = 0.0
    if gd.wasCanceled():  
        parm_Mode   = "Stop"
        IJ.log("Canceled")
        return str(parm_Mode), parm_Binary, parm_Gain, parm_Shadow,str(parm_Show), \
            parm_Prph, parm_FGGain, parm_FFCorrection, parm_FFC_sigma, parm_Gmixrate, \
            parm_subtractBv, parm_leveladj, parm_correct_fac, parm_HybridC, Fg_mask_type
    write_out(path, parm_Binary, parm_Gain, parm_Shadow, parm_Prph, \
            parm_FGGain, parm_FFCorrection, parm_FFC_sigma, parm_Gmixrate, \
            parm_subtractBv, parm_leveladj, parm_correct_fac, parm_HybridC, Fg_mask_type)
    return str(parm_Mode), parm_Binary, parm_Gain, parm_Shadow, str(parm_Show), \
        parm_Prph, parm_FGGain, parm_FFCorrection, parm_FFC_sigma, parm_Gmixrate, \
        parm_subtractBv, parm_leveladj, parm_correct_fac, parm_HybridC, Fg_mask_type
#----------------
def write_out(path, parm_Binary, parm_Gain, parm_Shadow, parm_Prph, \
            parm_FGGain, parm_FFCorrection, parm_FFC_sigma, parm_Gmixrate, \
            parm_subtractBv, parm_leveladj, parm_correct_fac, parm_HybridC, Fg_mask_type):
    wo_parm = dict(parm_Binary=parm_Binary, parm_Gain=parm_Gain, parm_Shadow=parm_Shadow, \
        parm_Prph=parm_Prph, parm_FGGain=parm_FGGain, parm_FFCorrection=parm_FFCorrection, \
        parm_FFC_sigma=parm_FFC_sigma, parm_Gmixrate=parm_Gmixrate, \
        parm_subtractBv=parm_subtractBv, parm_leveladj=parm_leveladj, \
        parm_correct_fac=parm_correct_fac, parm_HybridC=parm_HybridC, \
        Fg_mask_type=Fg_mask_type)
#  -- writing "filename_parameter.txt" file
    writing_file0 = open(path  + '_parameter.txt','a')
    # new_file_number = int(file_number) + 1
    # i = 0
    for key in wo_parm.keys():
        writing_file0.write(key + ": " + str(wo_parm[key]) +"\n")
    timenow = datetime.datetime.now()
    writing_file0.write("date time: " + timenow.strftime('%Y/%m/%d/%H:%M:%S') +"\n")
    writing_file0.close()
#-------------
def read_in(path):
    existing_parms = {}
    existing_parms["existing"] = "No"
        # define dic type V.
    is_file = os.path.isfile(path + '_parameter.txt')
    if is_file:
        # reading existing parameters 
        with open(path + '_parameter.txt','r') as read_file:
            lines = read_file.readlines()
            # IJ.showMessage(str(lines[0]))
            if len(lines) >= 14:
                for l in range(len(lines)-14, len(lines)-1):
                    eachline = lines[l].strip()
                    i = eachline.find(":")
                    key = eachline[:i]
                    value = eachline[i+2:]
                    existing_parms[key] = value
                    existing_parms["existing"] = "Yes"
    # print "existing: ", existing_parms["existing"], "Lines: ", len(lines)
    if existing_parms["existing"] == "No":
        existing_parms["parm_FFC_sigma"] = 20.0
        existing_parms["parm_subtractBv"] = "False"
        existing_parms["parm_leveladj"] = "Auto"
        existing_parms["parm_Shadow"] = 50.0
        existing_parms["parm_Prph"] = 10.0
        existing_parms["parm_correct_fac"] = 0.0
        existing_parms["parm_Binary"] = 128.0
        existing_parms["parm_Gmixrate"] = 0.0
        existing_parms["parm_Gain"] = 0.0
        existing_parms["parm_FFCorrection"] = "False"
        existing_parms["parm_FGGain"] = 128
        existing_parms["parm_HybridC"] = "Off"
        existing_parms["Fg_mask_type"] = "Type2"
    return existing_parms        
#-------------
# Main
od = OpenDialog("Choose a file")
folder = od.getDirectory()
file = od.getFileName()
if file:
    path = folder + file
    IJ.log(path)
    IJ.log(folder)
    IJ.log(file)
    parm = NoviceMenu(path)
    if parm[0] == "Expert":
        parm = getParameters(path)
    print ("Parm[0]=", parm[0])
    if parm[0] != "Stop":
        RGBmanipulation(folder,file,parm)
    IJ.log("Finish")
else:
    IJ.log("Job Canceled")
#    IJ.run("Quit")
